# -*- coding: utf-8 -*-
"""
Created on Mon Feb 11 11:39:38 2019

@author: famato1
"""
import itertools 

def countX(lst, x):
    '''An internal function counting the number of time an element is present in a list'''
    return lst.count(x) 

def list_combs(x):
    '''An internal function returning all the possible combination between n numbers'''
    return [c for i in range(len(x)+1) for c in itertools.combinations(x,i)]


def combs(x):
    '''An internal function chaining iterables'''
    return itertools.chain.from_iterable(
        itertools.combinations(x, i + 1)
        for i in range(len(x)+1))

def recursive_combination_excluder(lst, r, excludes):
    '''An internal function to recursively exclude combinations diring forward-backward feature selection'''
    return (comb for comb in itertools.combinations(lst, r)
                 if any(e.issubset(comb) for e in excludes))



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 16 10:30:37 2019

@author: famato1
"""
import numpy as np
from operator import itemgetter
from scipy import optimize
from sklearn.metrics import mean_squared_error as MSE
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.gaussian_process import kernels
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.model_selection import KFold

class GRNN(BaseEstimator, RegressorMixin):
    """A General Regression Neural Network, based on the Nadaraya-Watson estimator.
    Parameters:
    ----------
    kernel : str, default="rbf"
        Kernel function to be casted in the regression.
        The radial basis function (rbf) is used by default: K(x,x')=exp(-|x-x'|^2/(2 sigma^2))
        
    sigma : float, array, default=None
        Bandwidth standard deviation parameter for the Kernel. All Kernels are casted from the 
        sklearn.metrics.pairwise library. Check its documentation for further 
        info and to find a list of all the built-in Kernels.
    
    n_splits : int, default 10
        Number of folds used in the K-Folds CV definition of the sigma. 
        Must be at least 2.
    
    calibration : str, default=None
        Type of calibration of the sigma. 
        'gradient_search' minimizes the loss function by applying the scipy.optimize.minimize function. 
        Gradient search can be used for isotropic Kernels (sigma = int, float),
        or on anisotropic Kernel (sigma = list).
        'warm_start' is used when sigma is a single scalar. Gradient search is applied 
        to find the best sigma value for an isotropic Kernel (all sigmas are the same). The optimized 
        parameter is then used as a starting point to search the optimal solution for an 
        anisotropic Kernel (having one sigma per feature).
    
    method: str, default=L-BFGS-B
        Type of solver for the gradient search (used to find the local minimum of the cost function). 
        The default solver used is the Nelder-Mead. 
        Other choises (such as the CG based on the Polak and 
        Ribiere algorithm) are discussed on the help of the scipy function.
    
    bounds : list, default=(0, None)
        (min, max) pairs for each element in x, defining the bounds on that parameter.
        Use None or +-inf for one of min or max when there is no bound in that direction.
    
    n_restarts_optimizer : int, default = 0
        The number of restarts of the optimizer for finding the kernel's
        parameters which maximize the cost function. The first run
        of the optimizer is performed from the kernel's initial parameters,
        the remaining ones (if any) from inital sigmas sampled log-uniform randomly
        from the space of allowed sigma-values. 
    
    seed : int, default = 42
        Random state used to initialize random generators.   
    
    
    Notes
    -----
    This Python code was developed and used for the following papers:
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.

   
   References
    ----------
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.

    D.F. Specht. A general regression neural network. IEEE transactions on neural 
    networks 2.6 (1991): 568-576.


    Examples
    --------
    import numpy as np
    from sklearn import datasets
    from sklearn import preprocessing
    from sklearn.model_selection import train_test_split
    from sklearn.model_selection import  GridSearchCV
    from sklearn.metrics import mean_squared_error as MSE

    from pyGRNN import GRNN

    # Loading the diabetes dataset
    diabetes = datasets.load_diabetes()
    X = diabetes.data
    y = diabetes.target
    # Splitting data into training and testing
    X_train, X_test, y_train, y_test = train_test_split(preprocessing.minmax_scale(X),
                                                        preprocessing.minmax_scale(y.reshape((-1, 1))),
                                                        test_size=0.25)
    # Example 1: use Isotropic GRNN with a Grid Search Cross validation to select the optimal bandwidth
    IGRNN = GRNN()
    params_IGRNN = {'kernel':["RBF"],
                    'sigma' : list(np.arange(0.1, 4, 0.01)),
                    'calibration' : ['None']
                     }

    grid_IGRNN = GridSearchCV(estimator=IGRNN,
                              param_grid=params_IGRNN,
                              scoring='neg_mean_squared_error',
                              cv=5,
                              verbose=1
                              )
    grid_IGRNN.fit(X_train, y_train.ravel())
    best_model = grid_IGRNN.best_estimator_
    y_pred = best_model.predict(X_test)
    mse_IGRNN = MSE(y_test, y_pred)

    # Example 2: use Anisotropic GRNN with Limited-Memory BFGS algorithm to select the optimal bandwidths
    AGRNN = GRNN(calibration="gradient_search")
    AGRNN.fit(X_train, y_train.ravel())
    sigma=AGRNN.sigma 
    y_pred = AGRNN.predict(X_test)
    mse_AGRNN = MSE(y_test, y_pred)
    """

    def __init__(self, kernel='RBF', sigma=0.4, n_splits=5, calibration='warm_start', method='L-BFGS-B', bnds=(0, None), n_restarts_optimizer=0, seed = 42):
        self.kernel = kernel
        self.sigma = sigma
        self.n_splits = n_splits
        self.calibration = calibration
        self.method = method
        self.bnds = bnds
        self.n_restarts_optimizer = n_restarts_optimizer
        self.seed = seed
        
    def fit(self, X, y):
        """Fit the model.  
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The training input samples. Generally corresponds to the training features
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets
        Returns
        -------
        self : object
            Returns self.
        """

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        
        self.X_ = X
        self.y_ = y
        bounds = self.bnds
        
        np.seterr(divide='ignore', invalid='ignore')
        
        def cost(sigma_):
            '''Cost function to be minimized. It computes the cross validation
                error for a given sigma vector.'''
            kf = KFold(n_splits= self.n_splits, random_state=self.seed)
            kf.get_n_splits(self.X_)
            cv_err = []
            for train_index, validate_index in kf.split(self.X_):
                X_tr, X_val = self.X_[train_index], self.X_[validate_index]
                y_tr, y_val = self.y_[train_index], self.y_[validate_index]
                Kernel_def_= getattr(kernels, self.kernel)(length_scale=sigma_)
                K_ = Kernel_def_(X_tr, X_val)
                # If the distances are very high/low, zero-densities must be prevented:
                K_ = np.nan_to_num(K_)
                psum_ = K_.sum(axis=0).T # Cumulate denominator of the Nadaraya-Watson estimator
                psum_ = np.nan_to_num(psum_)
                y_pred_ = (np.dot(y_tr.T, K_) / psum_)
                y_pred_ = np.nan_to_num(y_pred_)
                cv_err.append(MSE(y_val, y_pred_.T))
            return np.mean(cv_err, axis=0) ## Mean error over the k splits                        
        
        def optimization(x0_):
            '''A function to find the optimal values of sigma (i.e. the values 
               minimizing the cost) given an inital guess x0.'''
            opt = optimize.minimize(cost, x0_, method=self.method, bounds=self.bnds)
            if opt['success'] is True:
                opt_sigma = opt['x']
                opt_cv_error = opt['fun']
            else:
                opt_sigma = np.full(len(self.X_[0]), np.nan)
                opt_cv_error = np.inf
                pass
            return [opt_sigma, opt_cv_error]
        
        def calibrate_sigma(self):
            '''A function to find the values of sigma minimizing the CV-MSE. The 
            optimization is based on scipy.optimize.minimize.'''    
            x0 = np.asarray(self.sigma) # Starting guess (either user-defined or measured with warm start)
            if self.n_restarts_optimizer > 0:
                #First optimize starting from theta specified in kernel
                optima = [optimization(x0)] 
                # # Additional runs are performed from log-uniform chosen initial bandwidths
                r_s = np.random.RandomState(self.seed)
                for iteration in range(self.n_restarts_optimizer): 
                    x0_iter = np.full(len(self.X_[0]), np.around(r_s.uniform(0,1), decimals=3))
                    optima.append(optimization(x0_iter))             
            elif self.n_restarts_optimizer == 0:        
                optima = [optimization(x0)]            
            else:
                raise ValueError('n_restarts_optimizer must be a positive int!')
            
            # Select sigma from the run minimizing cost
            cost_values = list(map(itemgetter(1), optima))
            self.sigma = optima[np.argmin(cost_values)][0]
            self.cv_error = np.min(cost_values) 
            return self
        
        
        if self.calibration is 'warm_start':
            print('Executing warm start...')
            self.bnds = (bounds,)           
            x0 = np.asarray(self.sigma)
            optima = [optimization(x0)]            
            cost_values = list(map(itemgetter(1), optima))
            self.sigma = optima[np.argmin(cost_values)][0]
            print('Warm start concluded. The optimum isotropic sigma is ' + str(self.sigma))
            self.sigma = np.full(len(self.X_[0]), np.around(self.sigma, decimals=3))
            self.bnds = (bounds,)*len(self.X_[0])
            #print ('Executing gradient search...')
            calibrate_sigma(self)
            print('Gradient search concluded. The optimum sigma is ' + str(self.sigma))
        elif self.calibration is 'gradient_search':
            #print ('Executing gradient search...')
            self.sigma = np.full(len(self.X_[0]), self.sigma)
            self.bnds = (bounds,)*len(self.X_[0])
            calibrate_sigma(self)
            #print('Gradient search concluded. The optimum sigma is ' + str(self.sigma))
        else:
            pass
                   
        self.is_fitted_ = True
        # Return the regressor
        return self
     
    def predict(self, X):
        """Predict target values for X.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the testing features
        Returns
        -------
        y : array of shape = [n_samples]
            The predicted target value.
        """
        
         # Check if fit had been called
        check_is_fitted(self, ['X_', 'y_'])
        
        # Input validation
        X = check_array(X)
        
        Kernel_def= getattr(kernels, self.kernel)(length_scale=self.sigma)
        K = Kernel_def(self.X_, X)
        # If the distances are very high/low, zero-densities must be prevented:
        K = np.nan_to_num(K)
        psum = K.sum(axis=0).T # Cumulate denominator of the Nadaraya-Watson estimator
        psum = np.nan_to_num(psum)
        return np.nan_to_num((np.dot(self.y_.T, K) / psum))



from distutils.core import setup
setup(
  name = 'pyGRNN',         
  packages = ['pyGRNN'],   
  version = 'v0.1',      
  license='MIT',
  description = 'Python implementation of General Regression Neural Network (Nadaraya-Watson Estimator). A Feature Selection module based on GRNN is also provided',   # Give a short description about your library
  author = 'Federico Amato',                   
  author_email = 'federico.amato@unil.ch',      #
  url = 'https://github.com/federhub/pyGRNN',  
  download_url = 'https://github.com/federhub/pyGRNN/archive/v0.1.tar.gz',    
  keywords = ['Machine Learning', 'General Regression Neural Network', 'Kernel Regression', 'Feature Selection'],   
  install_requires=[
          'pandas',
          'numpy',
          'seaborn',
          'scikit-learn',
          'matplotlib',
          'scipy',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',     
    'Intended Audience :: Developers',     
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3.6',
  ],
)



# -*- coding: utf-8 -*-
"""
Created on Fri Feb  8 15:03:06 2019

@author: famato1
"""

from .base import countX, combs, recursive_combination_excluder
from .GRNN import GRNN


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
from sklearn import preprocessing
import numpy as np
from sklearn.model_selection import  GridSearchCV
from sklearn.metrics import mean_squared_error as MSE
import itertools 


class Isotropic_selector():
    """
        A class to perform supervised feature selection using Isotropic
        GRNN. It also allows an (unsupervised) analysis of the relatidness of 
        the input space.
    Parameters
    ----------
    bandwidth : str, default="rule-of-thumb"
        The strategy through which the kernel bandwidth is choosen. Can be one 
        between the Silverman's rule of thumb ('rule-of-thumb') and the grid search
        Cross Validation ('GSCV'). 
    
    Notes
    -----
    This Python code was developed and used for the following papers:
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.


    References
    ----------
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.

    D.F. Specht. A general regression neural network. IEEE transactions on neural 
    networks 2.6 (1991): 568-576.


    Examples
    --------
    import numpy as np
    from sklearn import datasets
    from sklearn import preprocessing
    from sklearn.model_selection import train_test_split
    from sklearn.model_selection import  GridSearchCV
    from sklearn.metrics import mean_squared_error as MSE

    from pyGRNN import feature_selection as FS

    # Loading the diabetes dataset
    diabetes = datasets.load_diabetes()
    X = diabetes.data
    y = diabetes.target
    featnames = diabetes.feature_names

    # Example 1: use Isotropic Selector to explore data dependencies in the input 
    # space by analyzing the relatedness between features 
    IsotropicSelector = FS.Isotropic_selector(bandwidth = 'rule-of-thumb')
    IsotropicSelector.relatidness(X, feature_names = featnames)
    IsotropicSelector.plot_(feature_names = featnames)

    # Example 2: use Isotropic Selector to perform an exhaustive search; a rule-of-thumb
    # is used to select the optimal bandwidth for each subset of features
    IsotropicSelector = FS.Isotropic_selector(bandwidth = 'rule-of-thumb')
    IsotropicSelector.es(X, y.ravel(), feature_names=featnames)


    # Example 3: use Isotropic Selector to perform a complete analysis of the input
    # space, recongising relevant, redundant, irrelevant features
    IsotropicSelector = FS.Isotropic_selector(bandwidth = 'rule-of-thumb')
    IsotropicSelector.feat_selection(X, y.ravel(), feature_names=featnames, strategy ='es')
    """
    
    def __init__(self,bandwidth = 'rule-of-thumb'):
        self.bandwidth = bandwidth

    def isotropic_GSCV(X_tr, y_tr):
        '''Function to apply the Gridsearch CV to select the best bandwidth for 
            an isotropic GRNN.
        Parameters
        ----------
        X_tr : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y_tr : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.'''
        IGRNN = GRNN()
        params_IGRNN = {'sigma' : list(np.arange(0.01, 1, 0.01)),
                        'calibration' : ['None']}
        grid_IGRNN = GridSearchCV(estimator=IGRNN,
                                  param_grid=params_IGRNN,
                                  scoring='neg_mean_squared_error',
                                  cv=5,
                                  verbose=1,
                                  n_jobs=-1)
        grid_IGRNN.fit(X_tr, y_tr.ravel()) 
        return grid_IGRNN.best_estimator_
            
    def es(self, X, y, feature_names):
        '''Exhaustive search evaluation of the importance of each feature 
           using IGRNN;can be used to detect relevant features in the input space.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        '''
        self.ex_search = {}
        self.X = X
        self.y = y
        self.best_inSpaceNames = []
        #s = 0.9 * (self.X.shape[0])**-(1/5) # Silverman's rule of thumb, for std=1 (z-scored data)
        var = list(np.arange(0, self.X.shape[1], 1))
        series =combs(var)#[1::]
        print('Exploring the ' +str(2**len(var)-1) + ' possible combination of features...')
        # Set counter
        i = 0
        # Run a model for each computation
        self.mse_cv = 999999 #initalising a very large int variable
        for fset in series:
            #print ('Exploring combination ' + str(i) + '...') 
            X_run= self.X[:, list(fset)] 
            X_train, X_test, y_train, y_test = train_test_split(preprocessing.StandardScaler().fit_transform(X_run),
                                                                y.reshape((-1, 1)),
                                                                test_size=0.25,
                                                                random_state = 42)
            # searching the optimal sigma, depending on the strategy defined
            if self.bandwidth == 'rule-of-thumb':
                s = 0.9 * (X_run.shape[0])**-(1/(4+ X_run.shape[1])) # Silverman's rule of thumb, for std=1 (z-scored data) see book Li, Racine page 66
                IGRNN = GRNN(sigma = s, calibration = 'None')
                best_model = IGRNN.fit(X_train, y_train.ravel())
            elif self.bandwidth == 'GSCV':
                best_model = Isotropic_selector.isotropic_GSCV(X_train, y_train.ravel())
            else: 
                raise ValueError('Bandwidth search strategy must be one between rule-of-thumb and GSCV!')
            # Run a GRNN with the selected bandwidth kernel 
            y_pred = best_model.predict(X_test)
            self.ex_search.update({fset :[MSE(y_test, y_pred)]})
            if (MSE(y_test, y_pred)) < self.mse_cv:
                self.mse_cv = MSE(y_test, y_pred)
                self.best_inSpaceIndex = fset 
                self.best_inSpace = X_run
                self.best_model = best_model
            else:
                pass
            #print('Combination ' + str(i) +' tested.')
            i += 1   
        [self.best_inSpaceNames.append(feature_names[n]) for n in self.best_inSpaceIndex]
        print('The best subset of features is ' + str(self.best_inSpaceNames))
        return self  
  
    def ffs(self, X, y, feature_names, stop_criterion = 'first_min'):
        '''Forward feature selctor for the evaluation of the importance of each
           feature using IGRNN;can be used to detect relevant features in the input space.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        stop_criterion: str, default="first_min"
            The stopping criterion of the algorithm. If 'first_min', the search
            will end as soon as adding another feature to the best subset causes 
            an increse of the MSE of the model. Differently, when 'full_search' is 
            selected, the entire forward feature selection will be performed. 
        '''
        y_ = y
        n_feat = np.arange(0, X.shape[1],1)
        self.mse_cv = 999999 #initalising a very large int variable
        mse_run_best = 999999 #initalising a very large int variable
        self.best_inSpaceNames = []
        best_comb = [{f} for f in n_feat] # initializing a list of sets for the definition of combinations
        for i in n_feat+1:
            if self.bandwidth == 'rule-of-thumb':
                s = 0.9 * (X.shape[0])**-(1/(4+ i)) # Silverman's rule of thumb, for std=1 (z-scored data) see book Li, Racine page 66    
            else:
                pass
            for fset in recursive_combination_excluder(n_feat, i, best_comb):
                X_ = X[:, list(fset)]
                X_train, X_test, y_train, y_test = train_test_split(preprocessing.StandardScaler().fit_transform(X_),
                                                                    y_.reshape((-1, 1)),
                                                                    test_size=0.25,
                                                                    random_state = 42)
                        # searching the optimal sigma, depending on the strategy defined
                if self.bandwidth == 'rule-of-thumb':
                    IGRNN = GRNN(sigma = s, calibration = 'None')
                    best_model = IGRNN.fit(X_train, y_train.ravel())
                elif self.bandwidth == 'GSCV':
                    best_model = Isotropic_selector.isotropic_GSCV(X_train, y_train.ravel())
                else: 
                    raise ValueError('Bandwidth search strategy must be one between rule-of-thumb and GSCV!')
                # Run a GRNN with the selected bandwidth kernel 
                y_pred = best_model.predict(X_test)
                if (MSE(y_test, y_pred)) < mse_run_best:
                    mse_run_best = MSE(y_test, y_pred)
                    best_comb_run = fset
                    X_run = X_
                else:
                    pass 
            if mse_run_best < self.mse_cv:
                self.best_inSpaceIndex = best_comb_run
                self.best_inSpace = X_run
                self.best_model = best_model
                self.mse_cv = mse_run_best   
                best_comb = [set(best_comb_run)]
            else:
                if stop_criterion == 'first_min':
                    break  
                else:
                    pass
        [self.best_inSpaceNames.append(feature_names[n]) for n in self.best_inSpaceIndex]
        return self, print('Best subset is: ' + str(self.best_inSpaceNames))
            
    def bfs(self, X, y, feature_names, stop_criterion = 'first_min'):
        '''Backward feature selctor for the evaluation of the importance of each
           feature using IGRNN;can be used to detect relevant features in the input space.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        stop_criterion: str, default="first_min"
            The stopping criterion of the algorithm. If 'first_min', the search
            will end as soon as adding another feature to the best subset causes 
            an increse of the MSE of the model. Differently, when 'full_search' is 
            selected, the entire forward feature selection will be performed.     
        '''
        y_ = y
        n_feat = -np.sort(-(np.arange(0, X.shape[1],1)))
        self.mse_cv = 999999 #initalising a very large int variable
        mse_run_best = 999999 #initalising a very large int variable
        self.best_inSpaceNames = []
        best_comb = ((f) for f in n_feat) # initializing a list of sets for the definition of combinations
        for i in n_feat+1:
            if self.bandwidth == 'rule-of-thumb':
                s = 0.9 * (X.shape[0])**-(1/(4+ i)) # Silverman's rule of thumb, for std=1 (z-scored data) see book Li, Racine page 66
            else:
                pass
            for fset in itertools.combinations(best_comb , i):            
                X_ = X[:, list(fset)]
                X_train, X_test, y_train, y_test = train_test_split(preprocessing.StandardScaler().fit_transform(X_),    
                                                                    y_.reshape((-1, 1)),
                                                                    test_size=0.25,
                                                                    random_state = 42)
                # searching the optimal sigma, depending on the strategy defined
                if self.bandwidth == 'rule-of-thumb':
                    IGRNN = GRNN(sigma = s, calibration = 'None')
                    best_model = IGRNN.fit(X_train, y_train.ravel())
                elif self.bandwidth == 'GSCV':
                    best_model = Isotropic_selector.isotropic_GSCV(X_train, y_train.ravel())
                else: 
                    raise ValueError('Bandwidth search strategy must be one between rule-of-thumb and GSCV!')
                # Run a GRNN with the selected bandwidth kernel 
                y_pred = best_model.predict(X_test)
                if (MSE(y_test, y_pred)) < mse_run_best:
                    mse_run_best = MSE(y_test, y_pred)
                    best_comb_run = fset
                    X_run = X_
                else:
                    pass   
            if mse_run_best < self.mse_cv:
                best_comb = ((f) for f in best_comb_run)
                self.best_inSpaceIndex = best_comb_run
                self.best_inSpace = X_run
                self.best_model = best_model
                self.mse_cv = mse_run_best  
                
            else:
                if stop_criterion == 'first_min':
                    break  
                else:
                    pass
        [self.best_inSpaceNames.append(feature_names[n]) for n in self.best_inSpaceIndex]
        return self, print('Best subset is: ' + str(self.best_inSpaceNames))

    
    def feat_selection(self, X, y, feature_names, strategy = 'ffs', stop_criterion = 'first_min'):
        ''' Feature selector based on IGRNN. For a given dataset, a complete
            exploration of the feature space is provided, giving information 
            about the relevance, irrelevance or the redundancy of the features.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        strategy : str, default = 'ffs'
            The search strategy for the feature selection. Can be forward ('ffs'),
            backward ('bfs') or exhaustive search ('es'). 
        stop_criterion: str, default="first_min"
            The stopping criterion of the algorithm. If 'first_min', the search
            will end as soon as adding another feature to the best subset causes 
            an increse of the MSE of the model. Differently, when 'full_search' is 
            selected, the entire forward feature selection will be performed.   
        '''
        in_feat = X
        output = y
        self.redundant = [] # we will store here the indexes of redundant features
        self.irrelevant = [] # we will store here the indexes of irrelevant features
        print('Searching relevant features...')
        if strategy is 'ffs':
            Isotropic_selector.ffs(self, in_feat, output, feature_names, stop_criterion = stop_criterion) # performing a forward feature selection to find relevant features
        elif strategy is 'bfs':
            Isotropic_selector.bfs(self, in_feat, output, feature_names, stop_criterion = stop_criterion) # performing backward feature selection to find relevant features
        else:
             Isotropic_selector.es(self, in_feat, output, feature_names) # performing an exhaustive search to find relevant features
        best__model_ = self.best_model 
        self.relevant = list(self.best_inSpaceIndex) # creating a list with the indexes of relevant features
        dataframe = pd.DataFrame(data=in_feat, columns=feature_names) # dataframe from the feature space
        for col in self.relevant:
            print('Searching the best subset to predict feature ' + str(col))
            feat = dataframe.drop(dataframe.columns[col], axis = 1).values
            target = dataframe.iloc[:,col].values
            # Getting the right names for the features
            feat_idx = [column for column in np.arange(0, in_feat.shape[1],1) if column != col] 
            feat_names = [feature_names[n] for n in feat_idx]
            # performing the search strategy
            if strategy is 'ffs':
                Isotropic_selector.ffs(self, feat, target, feat_names) # performing a forward feature selection to find relevant features
            elif strategy is 'bfs':
                Isotropic_selector.bfs(self, feat, target, feat_names) # performing backward feature selection to find relevant features
            else:
                Isotropic_selector.es(self, feat, target, feat_names) # performing an exhaustive search to find relevant features
            best_set_name = [feat_names[n] for n in  list(self.best_inSpaceIndex)] # getting the names of the best subset to predict the considered col 
            best_set_idx = [feature_names.index(m) for m in best_set_name] # getting the index of the best subset to predict the considered col 
            # usinf a (complicated) list cmprehension to list the feature which are
            # not in the best subset for the considered col
            not_relevant = [x for i,x in enumerate(list((range(0, in_feat.shape[1],1)))) if i not in best_set_idx] 
            print(not_relevant)
            for i in not_relevant: # if a feature cannot be used to predict a relevant feature, and it is not redundant for other relevant features, then it is irrelevant
                if i in self.relevant:
                    pass
                else:
                    if i in self.redundant:
                        pass
                    else:
                        if i not in self.irrelevant:
                            self.irrelevant.append(i)
                        else:
                            pass
            for j in best_set_idx: # if a feature can predict a relevant feature, it is redundant
                if j in self.relevant:
                    pass
                else:
                    if j in self.irrelevant:
                        self.irrelevant.remove(j)
                        if j not in self.redundant:
                            self.redundant.append(j)
                        else:
                            pass
                    else:
                        if j not in self.redundant:
                            self.redundant.append(j)
                        else:
                            pass   
        self.best_inSpaceIndex = self.relevant
        self.best_inSpace = X[:, self.relevant]
        self.best_model  = best__model_
        print('Research completed!')
        print('The relevant features are: ' + str([feature_names[n] for n in self.relevant]))    
        print('The redundant features are: ' + str([feature_names[n] for n in self.redundant]))   
        print('The irrelevant features are: ' + str([feature_names[n] for n in self.irrelevant]))   
        return self  
    
    def relatidness(self, X, feature_names, strategy = 'ffs'):
            '''Exhaustive search evaluation of the importance of each feature 
            in the input space using IGRNN; can be used to study the relatidness
            among the features in input or output space
            Parameters
            ----------
            X : array-like of shape = [n_samples, n_features]
                The input samples. Generally corresponds to the training features.
            feature_names : list
                A list containing the names of the input features.
            strategy : str, default = 'ffs'
                The search strategy for the feature selection. Can be forward ('ffs'),
                backward ('bfs') or exhaustive search ('es').              
            '''
            self.relatidness_ = {}
            dataframe = pd.DataFrame(data=X, columns=feature_names)
            #Defining a list of all the possible permutation of features
            for col in range(0,X.shape[1],1):
                print('Searching relatidness for feature ' + str(col))
                feat = dataframe.drop(dataframe.columns[col], axis = 1).values
                target = dataframe.iloc[:,col].values
                # Getting the right names for the features
                feat_idx = [column for column in np.arange(0, X.shape[1],1) if column != col] 
                feat_names = [feature_names[n] for n in feat_idx]
                # performing the search strategy
                if strategy is 'ffs':
                    Isotropic_selector.ffs(self, feat, target, feat_names) # performing a forward feature selection to find relevant features
                elif strategy is 'bfs':
                    Isotropic_selector.bfs(self, feat, target, feat_names) # performing backward feature selection to find relevant features
                else:
                    Isotropic_selector.es(self, feat, target, feat_names) # performing an exhaustive search to find relevant features
                self.relatidness_.update({col :[ [self.best_inSpaceNames]]})
            return self  
    
    def plot_(self, feature_names):
        '''A function to plot the relatidness matrix. Isotropic_selector.relatidness 
            has to be executed first.
         Parameters
            ----------
         feature_names : list
             A list containing the names of the input features.              
        '''
        if not self.relatidness_:
            raise ValueError("To plot results, Isotropic_selector.relatidness has to be executed first!")
        else:
            pass
        df = pd.DataFrame(columns = feature_names)
        feature = []
        best_set = []
        for k,v in self.relatidness_.items():
            feature.append(k)
            best_set.append(list(v[0][0]))
        col = 0
        for i in best_set: 
            presence = []
            for feat in feature_names:
                a = countX(i, feat)
                presence.append(a)
            df.iloc[:,col] = pd.Series(presence).values
            col += 1
        # plotting an heatmap with the best results
        plt.style.use('ggplot')
        with sns.axes_style("white"):
                f, ax = plt.subplots(figsize=(11, 9))
                ax.tick_params(labelsize = 14)
                ax.xaxis.set_ticks_position('top')
                cmap = sns.diverging_palette(220, 10, as_cmap=True)
                # Manually specify colorbar labelling after it's been generated
                p2 = sns.heatmap(df,cmap= cmap, square=True, linewidths=.5,
                                 yticklabels = feature_names, cbar_kws={"boundaries": (-1, 1, 2)})
                colorbar = ax.collections[0].colorbar
                colorbar.set_ticks([0, 1])
                colorbar.set_ticklabels(['Excluded', 'Included'])
        return p2                 
    
 
   
class Anisotropic_selector():
    """
        A class to perform supervised feature selection using Ansotropic
        GRNN. 
    Notes
    -----
    This Python code was developed and used for the following papers:
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.

    References
    ----------
    F. Amato, F. Guignard, P. Jacquet, M. Kanveski. Exploration of data dependencies
    and feature selection using General Regression Neural Networks.

    D.F. Specht. A general regression neural network. IEEE transactions on neural 
    networks 2.6 (1991): 568-576.


    Example
    --------
    import numpy as np
    from sklearn import datasets
    from sklearn import preprocessing
    from sklearn.model_selection import train_test_split
    from sklearn.model_selection import  GridSearchCV
    from sklearn.metrics import mean_squared_error as MSE

    from pyGRNN import feature_selection as FS

    # Loading the diabetes dataset
    diabetes = datasets.load_diabetes()
    X = diabetes.data
    y = diabetes.target
    featnames = diabetes.feature_names

    # Using Anisotropic Selector to exclude irrelevant features
    AnisotropicSelector = FS.Anisotropic_selector()
    AnisotropicSelector.max_dist(X, y.ravel(), feature_names=featnames)
    """
    
    def bfe(self, X, y, feature_names, strategy = 'ffs',  stop_criterion = 'first_min', n_restarts_optimizer=0):
        '''Backward feature selctor for the evaluation of the importance of each
           feature using AGRNN;can be used to detect relevant features in the input space.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        stop_criterion: str, default="first_min"
            The stopping criterion of the algorithm. If 'first_min', the search
            will end as soon as adding another feature to the best subset causes 
            an increse of the MSE of the model. Differently, when 'full_search' is 
            selected, the entire forward feature selection will be performed.
        n_restarts_optimizer : int, default = 0
            The number of restarts of the optimizer for finding the kernel's
            parameters which maximize the cost function. The first run
            of the optimizer is performed from the kernel's initial parameters,
            the remaining ones (if any) from inital sigmas sampled log-uniform randomly
            from the space of allowed sigma-values. 
        '''
        self.mse_cv = 999999 #initalising a very large int variable
        y_ = y
        self.best_inSpaceNames = []
        self.best_inSpaceIndex = list(np.arange(0, X.shape[1],1))
                
        for maxiter in range(0, X.shape[1]):
            X_ = X[:, self.best_inSpaceIndex]
            X_train, X_test, y_train, y_test = train_test_split(X_,
                                                                y_.reshape((-1, 1)),
                                                                test_size=0.25,
                                                                random_state = 42)
            AGRNN = GRNN(calibration="warm_start", method="L-BFGS-B", n_restarts_optimizer=n_restarts_optimizer)
            best_model = AGRNN.fit(X_train, y_train.ravel())
            mse_run = MSE(y_test, best_model.predict(X_test))
            print('the current mse is '+str(mse_run))
            if mse_run < self.mse_cv:
                del self.best_inSpaceIndex[np.argmax(best_model.sigma)]
                self.best_inSpace = X_
                self.best_model = best_model
                self.mse_cv = mse_run        
            else :
                 if stop_criterion == 'first_min':
                    break  
                 else:
                    pass
        [self.best_inSpaceNames.append(feature_names[n]) for n in self.best_inSpaceIndex]
        return self, print('Best subset is: ' + str(self.best_inSpaceNames))
    
    def max_dist(self, X, y, feature_names, n_restarts_optimizer=0):
        '''Feature selctor for the evaluation of relevant feature using AGRNN.
           Can be used to detect relevant features in the input space. All the
           features which have bandwidth value lower than sqrt(n_feat) after 
           the model calibration are labeled as relevant.
        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples. Generally corresponds to the training features.
        y : array-like, shape = [n_samples]
            The output or target values. Generally corresponds to the training targets.
        feature_names : list
            A list containing the names of the input features.
        n_restarts_optimizer : int, default = 0
            The number of restarts of the optimizer for finding the kernel's
            parameters which maximize the cost function. The first run
            of the optimizer is performed from the kernel's initial parameters,
            the remaining ones (if any) from inital sigmas sampled log-uniform randomly
            from the space of allowed sigma-values. 
        '''
        AGRNN = GRNN(calibration="warm_start", method="L-BFGS-B", n_restarts_optimizer=n_restarts_optimizer)
        AGRNN.fit(X, y.ravel())
        self.best_inSpaceIndex = [i for i in range(len(AGRNN.sigma)) if AGRNN.sigma[i] < np.sqrt(len(AGRNN.sigma))]
        self.best_inSpaceNames = [feature_names[n] for n in self.best_inSpaceIndex]
        self.best_inSpace = X[:, self.best_inSpaceIndex]
        return self, print('Best subset is: ' + str(self.best_inSpaceNames))
